// Crit-bit trees

#include <stdio.h>  // for cbt_print
#include <stdlib.h>  // for NULL
#include <string.h>  // for strcmp
#include "cbt.h"

enum cbt_type {
    cbt_node,
    cbt_leaf,
};

static inline int chartestbit(char c, int bit)
{
    return (1 << (7 - bit)) & c;
}

static int firstcritbit(int *critbit, char *key0, char *key1)
{
    char *cp0 = key0;
    char *cp1 = key1;
    int bit;

    while(*cp0 == *cp1) {
	if (!*cp0) {
	    return 0;
	}
	cp0++;
	cp1++;
    }

    char c = *cp0 ^ *cp1;
    for (bit = 0; !chartestbit(c, bit); bit++);
    *critbit = ((cp0 - key0) << 3) + bit;
    if (chartestbit(*cp0, bit)) return 1;
    return -1;
}

static void cbt_node_init(cbt_node_ptr n)
{
    n->parent = NULL;
}

static cbt_node_ptr new_cbt_node()
{
    cbt_node_ptr p = malloc(sizeof(cbt_node_t));
    cbt_node_init(p);
    return p;
}

static int testbit(char *key, int bit)
{
    int i = bit >> 3;
    return chartestbit(key[i], (bit & 7));
}

static cbt_node_ptr cbt_leaf_at(cbt_node_ptr n, char *key)
{
    cbt_node_ptr p = n;
    int len = (strlen(key) << 3) - 1;
    for (;;) {
	if (cbt_leaf == p->type) {
	    return p;
	}
	if (len < p->critbit) {
	    do {
		p = p->left;
	    } while (cbt_node == p->type);
	    return p;
	}
	if (testbit(key, p->critbit)) {
	    p = p->right;
	} else {
	    p = p->left;
	}
    }
}

void *cbt_at(cbt_ptr cbt, char *key)
{
    if (!cbt->root) return NULL;
    cbt_node_ptr p = cbt_leaf_at(cbt->root, key);
    if (!strcmp(p->key, key)) {
	return p->data;
    }
    return NULL;
}

static void cbt_node_put_leaf(cbt_node_ptr n, void *data, char *key)
{
    n->type = cbt_leaf;
    n->data = data;
    n->key = key;
}

void cbt_put(cbt_ptr cbt, void *data, char *key)
{
    if (!cbt->root) {
	cbt->root = new_cbt_node();
	cbt_node_put_leaf(cbt->root, data, key);
	return;
    }
    cbt_node_ptr p = cbt_leaf_at(cbt->root, key);
    if (!strcmp(p->key, key)) {
	p->data = data;
	return;
    }
    cbt_node_ptr pleaf = new_cbt_node();
    cbt_node_ptr pnode = new_cbt_node();
    cbt_node_put_leaf(pleaf, data, key);
    int res = firstcritbit(&pnode->critbit, key, p->key);
    while (p->parent && (p->parent->critbit > pnode->critbit)) {
	p = p->parent;
    }
    if (!p->parent) {
	cbt->root = pnode;
    } else {
	if (p->parent->left == p) {
	    p->parent->left = pnode;
	} else {
	    p->parent->right = pnode;
	}
	pnode->parent = p->parent;
    }
    p->parent = pnode;
    pleaf->parent = pnode;
    pnode->type = cbt_node;

    if (res < 0) {
	pnode->left = pleaf;
	pnode->right = p;
    } else {
	pnode->left = p;
	pnode->right = pleaf;
    }
}

void cbt_init(cbt_ptr cbt)
{
    cbt->root = NULL;
}

static void cbt_node_print(cbt_node_ptr n)
{
    if (cbt_leaf == n->type) {
	printf("\"%s\"", n->key);
    } else {
	printf("[%d, ", n->critbit);
	cbt_node_print(n->left);
	printf(", ");
	cbt_node_print(n->right);
	printf("]");
    }
}

void cbt_print(cbt_ptr cbt)
{
    if (!cbt->root) {
	printf("empty\n");
	return;
    }
    cbt_node_print(cbt->root);
    printf("\n");
}

void *cbt_alloc()
{
    return malloc(sizeof(struct cbt_s));
}
