// cuckoo hash table

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "hashtab.h"

void hashtab_init(hashtab_ptr h)
{
    h->count = 0;
    h->array_size = 11;
    h->array = malloc(h->array_size * sizeof(hashtab_entry_ptr));
    memset(h->array, 0, h->array_size * sizeof(hashtab_entry_ptr));
}

static void compute_hashes(int *hash1p, int *hash2p, hashtab_ptr h, char *key)
{
    char *c;

    unsigned int hash = 0;
    unsigned int hash2 = 0;
     
    // from http://en.wikipedia.org/wiki/Hash_table
    // (algorithm due to Bob Jenkins)
    for (c = key; *c; c++) {
	hash += *c;
	hash += (hash << 10);
	hash ^= (hash >> 6);
    }

    hash += (hash << 3);
    hash ^= (hash >> 11);
    hash += (hash << 15);

    hash2 = hash % h->array_size;
    hash >>= 16;
    hash %= h->array_size;
    if (hash == hash2) {
	hash2++;
	hash2 %= h->array_size;
    }
    *hash1p = hash;
    *hash2p = hash2;
}

void hashtab_print(hashtab_ptr h)
{
    int i;
    for (i = 0; i < h->array_size; i++) {
	hashtab_entry_ptr p;
	p = h->array[i];
	if (p) {
	    int hash1, hash2;
	    compute_hashes(&hash1, &hash2, h, p->key);
	    printf("%d (%d/%d): %s\n", i, hash1, hash2, p->key);
	}
    }
}

static int hashtab_internal_put(hashtab_ptr h, void *data, char *key)
{
    int hash1, hash2;
    compute_hashes(&hash1, &hash2, h, key);
    hashtab_entry_ptr p1, p2;
    p1 = h->array[hash1];
    p2 = h->array[hash2];
    if (!p1) {
	p1 = malloc(sizeof(hashtab_entry_t));
	p1->key = key;
	p1->data = data;
	h->count++;
	h->array[hash1] = p1;
	return 0;
    } else if (!strcmp(p1->key, key)) {
	p1->data = data;
	return 0;
    } else if (!p2) {
	p2 = malloc(sizeof(hashtab_entry_t));
	p2->key = key;
	p2->data = data;
	h->array[hash2] = p2;
	h->count++;
	return 0;
    } else if (!strcmp(p2->key, key)) {
	p2->data = data;
	return 0;
    } else {
	// TODO: randomly bump one of the two locations?
	// right now we bump the second all the time
	int bumpedi = hash2;
	p1 = malloc(sizeof(hashtab_entry_t));
	p1->key = key;
	p1->data = data;
	h->count++;
	for(;;) {
	    p2 = h->array[bumpedi];
	    h->array[bumpedi] = p1;
	    if (!p2) return 0;
	    compute_hashes(&hash1, &hash2, h, p2->key);
	    if (hash2 == bumpedi) {
		bumpedi = hash1;
	    } else {
		bumpedi = hash2;
	    }
	    p1 = p2;
	    if (!strcmp(p1->key, key)) {
		printf("cycle %p\n", p1);
		free(p1);
		return 1;
	    }
	}
    }
    return 0;
}

static void hashtab_double(hashtab_ptr h)
{
    int i;
    int oldsize = h->array_size;
    hashtab_entry_ptr *oldarray = h->array;
    h->array_size = 2 * oldsize + 1;
    h->array = malloc(h->array_size * sizeof(hashtab_entry_ptr));
    memset(h->array, 0, h->array_size * sizeof(hashtab_entry_ptr));
    for (i = 0; i < oldsize; i++) {
	hashtab_entry_ptr p;
	p = oldarray[i];
	if (p) {
	    if (hashtab_internal_put(h, p->data, p->key)) {
		printf("clear h->array, double it, try again\n");
		exit(1);
	    }
	}
    }
    for (i = 0; i < oldsize; i++) {
	free(oldarray[i]);
    }
    free(oldarray);
}

void hashtab_put(hashtab_ptr h, void *data, char *key)
{
    if (h->count * 2 + 1 >= h->array_size) {
	hashtab_double(h);
    }
    while (hashtab_internal_put(h, data, key)) {
	hashtab_double(h);
    }
}

void *hashtab_at(hashtab_ptr h, char *key)
{
    hashtab_entry_ptr p1, p2;
    int hash1, hash2;
    compute_hashes(&hash1, &hash2, h, key);
    p1 = h->array[hash1];
    if (p1 && !strcmp(p1->key, key)) {
	return p1->data;
    }
    p2 = h->array[hash2];
    if (p2 && !strcmp(p2->key, key)) {
	return p2->data;
    }
    return NULL;
}

void *hashtab_alloc()
{
    return malloc(sizeof(struct hashtab_s));
}
